#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

import argparse

from transmedia.conversion import (ConvertBytesInFileToPngFile,
                                   ConvertPixelsInPngFileToBytesFile)

from transmedia.util import calculate_output_png_width_from_file


if __name__ == '__main__':
    # Parse arguments
    parser = argparse.ArgumentParser(description='Transform bytes into PNG or vice versa')
    parser.add_argument('-i', '--input', help='input file', required=True)
    parser.add_argument('-o', '--output', help='output file', required=True)
    args = parser.parse_args()

    if args.input.endswith(".png"):
        ConvertPixelsInPngFileToBytesFile(args.input, args.output).execute()

    elif args.output.endswith(".png"):
        png_width = calculate_output_png_width_from_file(args.input)
        ConvertBytesInFileToPngFile(args.input, args.output, png_width).execute()
    else:
        raise ValueError("Either the input or output file must end in '.png'")



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

import io

from transmedia.util import PixelArray
from transmedia.io import (write_data_to_file, write_pixels_to_png)


def test_write_png_produces_file_of_expected_size():
    expected_size = 77  # Precomputed size of resulting PNG
    white = [255, 255, 255]
    black = [0, 0, 0]
    pixel_rows = [black + white + black,
                  white + black + white,
                  black + white + black]

    pixel_array = PixelArray(pixel_rows)
    file = io.BytesIO()
    write_pixels_to_png(pixel_array, file)

    assert len(file.getvalue()) == expected_size


def test_write_data_produces_expected_output():
    data = bytearray.fromhex('ffffe3990012')
    file = io.BytesIO()
    write_data_to_file(data, file)

    assert file.getvalue() == data



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

import png

from transmedia.util import PixelArray


def read_data_from_file(file):
    """Read all data from a file into a bytearray.

    :param file: file open for binary read
    :type file: io.BufferedReader() or io.BytesIO()

    :return: bytearray of the provided file
    :rtype: bytes()
    """
    return b''.join([b for b in file])


def write_data_to_file(data, file):
    """Write data to a file.

    :param data: a bytearray
    :type data: bytes()
    :param file: target file
    :type file: io.BufferedReader() or io.BytesIO()
    """
    file.write(data)


def read_pixels_from_png(file):
    """Read pixels from a PNG file into a pixel array object.

    :param file: file open for binary read
    :type file: io.BufferedReader() or io.BytesIO()

    :return: pixel array object
    :rtype: PixelArray()
    """
    png_data = png.Reader(file).read()
    pixel_imap = png_data[2]

    return PixelArray([r.tolist() for r in pixel_imap])


def write_pixels_to_png(pixel_array, file):
    """Write rows of flat pixels to a png file.

    :param pixel_array: a pixel array object
    :type pixel_array: PixelArray()
    :param file: target png file
    :type file: io.BufferedReader() or io.BytesIO()
    """
    w = png.Writer(pixel_array.width, pixel_array.height, compression=9)
    w.write(file, pixel_array.rows)



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'


def test_import_of_all_modules():
    from transmedia import conversion
    from transmedia import encoders
    from transmedia import io
    from transmedia import util
    from transmedia import transform_png



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

import transmedia.encoders as encoders


### simple_encode tests ###

def check_simple_encode_maps_word_to_correct_color(word, correct_color):
    test_color = encoders.simple_encode(word)

    assert test_color == correct_color, \
        "{} =! {}".format(test_color, correct_color)


def test_simple_encode_maps_words_to_correct_colors():
    values = ((bytearray.fromhex('0000'), (int(0x00), int(0x00), int(0x00))),
              (bytearray.fromhex('1111'), (int(0x11), int(0x11), int(0x00))),
              (bytearray.fromhex('ffff'), (int(0x00), int(0x01), int(0xff))))

    for word, correct_color in values:
        yield(check_simple_encode_maps_word_to_correct_color, word,
              correct_color)


def test_simple_encode_raises_value_error_on_single_byte_input():
    try:
        encoders.simple_encode(bytearray.fromhex('00'))

        assert False, "simple_encoder erroneously succeeds on single-byte input"

    except ValueError:
        pass


def test_simple_encode_raises_type_error_on_int_input():
    try:
        encoders.simple_encode(1)

        assert False, "simple_encoder erroneously succeeds on integer input"

    except TypeError:
        pass


### simple_decode tests ###

def check_simple_decode_maps_color_to_correct_word(color, correct_word):
    test_word = encoders.simple_decode(*color)

    assert test_word == correct_word, "{} =! {}".format(test_word, correct_word)


def test_simple_decode_maps_colors_to_correct_words():
    values = ((bytearray.fromhex('0000'), (int(0x00), int(0x00), int(0x00))),
              (bytearray.fromhex('1111'), (int(0x11), int(0x11), int(0x00))),
              (bytearray.fromhex('ffff'), (int(0x00), int(0x01), int(0xff))))

    for correct_word, color in values:
        yield(check_simple_decode_maps_color_to_correct_word, color,
              correct_word)


def test_simple_decode_raises_type_error_on_single_byte_input():
    try:
        encoders.simple_decode(0)

        assert False, "simple_decoder erroneously succeeds on single-byte input"

    except TypeError:
        pass


def test_simple_decode_raises_type_error_on_str_input():
    try:
        encoders.simple_decode('0', '0', '0')

        assert False, "simple_decoder erroneously succeeds on string input"

    except TypeError:
        pass



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

from transmedia import encoders
from transmedia.io import (read_data_from_file, write_pixels_to_png,
                           read_pixels_from_png, write_data_to_file)

from transmedia.util import PixelArray


def convert_bytes_to_pixels(data, width, encoder=encoders.simple_encode):
    """Given a target pixel width and encoding function, convert bytes to
    rows of pixels, two bytes at a time.

    :param data: bytes to encode
    :type data: iterable
    :param width: number of pixels per row
    :type width: int
    :param encoder: encoding function
    :type encoder: function()

    :return: a pixel array object
    :rtype: PixelArray()
    """
    if width < 1:
        raise ValueError("width parameter must be a positive integer")

    byte_depth = 3
    bytes_per_pixel = 2
    data_idx = 0
    rows = []
    try:
        while True:
            cur_row = []
            while len(cur_row) < (width * byte_depth):
                word = data[data_idx:data_idx + bytes_per_pixel]
                if len(word) < bytes_per_pixel:
                    if len(cur_row) > 0:
                        rows += [cur_row]

                    raise ValueError()

                cur_row += encoder(word)
                data_idx += bytes_per_pixel

            rows += [cur_row]

    except ValueError:
        pass

    return PixelArray(rows, byte_depth=byte_depth)


def convert_pixels_to_bytes(pixel_array, encoder=encoders.simple_decode):
    """Given a pixel array and an encoding function, convert all pixels
    to a byte array.

    :param pixel_array: a pixel array object
    :type pixel_array: PixelArray()
    :param encoder: encoding function
    :type encoder: function()

    :return: a bytearray
    :rtype: bytes()
    """
    words = bytearray()
    for row in pixel_array:
        byte_idx = 0
        while byte_idx < len(row) - 1:
            try:
                components = row[byte_idx:byte_idx + pixel_array.byte_depth]
                words += encoder(*components)
                byte_idx += pixel_array.byte_depth

            except ValueError:
                break

    return words


class ConversionBase():
    """Base class for all conversion objects.
    """
    def __init__(self, input_file, output_file):
        """Initialize object.

        :param input_file: path of the file to convert
        :type input_file: str
        :param output_file: path of the file to be written
        :type output_file: str
        """
        self._input_file = input_file
        self._output_file = output_file

    def execute(self):
        raise NotImplementedError("execute() method not implemented!")


class ConvertBytesInFileToPngFile(ConversionBase):
    def __init__(self, input_file, output_file, width):
        """Initialize object.

        :param input_file: path of the file to convert
        :type input_file: str
        :param output_file: path of the file to be written
        :type output_file: str
        :param width: width, in pixels, of the resulting PNG image
        :type width: int
        """
        super().__init__(input_file, output_file)
        self._width = width

    def execute(self):
        with open(self._input_file, 'rb') as data_file:
            with open(self._output_file, 'wb') as png_file:
                data = read_data_from_file(data_file)
                write_pixels_to_png(convert_bytes_to_pixels(data, self._width),
                                    png_file)


class ConvertPixelsInPngFileToBytesFile(ConversionBase):
    def __init__(self, input_file, output_file):
        """Initialize object.

        :param input_file: path of the file to convert
        :type input_file: str
        :param output_file: path of the file to be written
        :type output_file: str
        """
        super().__init__(input_file, output_file)

    def execute(self):
        with open(self._input_file, 'rb') as png_file:
            with open(self._output_file, 'wb') as data_file:
                pixels = read_pixels_from_png(png_file)
                write_data_to_file(convert_pixels_to_bytes(pixels), data_file)




#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

from transmedia.conversion import (convert_bytes_to_pixels,
                                   convert_pixels_to_bytes)

import transmedia.encoders as encoders


### bytes2pixels tests ###

def check_bytes2pixels_returns_correct_width(correct_width):
    data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
    pixels = convert_bytes_to_pixels(data, correct_width,
                                     encoders.simple_encode)

    width_top = len(pixels.rows[0]) / pixels.byte_depth
    width_bottom = len(pixels.rows[-1]) / pixels.byte_depth

    assert width_top == width_bottom == correct_width, \
        "({} == {} == {}) is false".format(width_top, width_bottom,
                                           correct_width)


def test_bytes2pixels_returns_correct_width():
    for width in (1, 3, 9):  # a single column, a square, a single row
        yield check_bytes2pixels_returns_correct_width, width


def test_bytes2pixels_returns_square():
    correct_dimension = 3  # 3x3 square
    data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
    pixels = convert_bytes_to_pixels(data, correct_dimension,
                                     encoders.simple_encode)

    width_top = len(pixels.rows[0]) / pixels.byte_depth
    height = len(pixels.rows)

    assert width_top == height == correct_dimension, \
        "({} == {} == {} == {}) is false".format(width_top, height,
                                                 correct_dimension)


def test_bytes2pixels_raises_value_error_on_negative_width():
    try:
        negative_width = -1
        data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
        convert_bytes_to_pixels(data, negative_width, encoders.simple_encode)

        assert False, "bytes2pixels erroneously succeeds on negative width"

    except ValueError:
        pass


def test_bytes2pixels_raises_value_error_on_zero_width():
    try:
        zero_width = 0
        data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
        convert_bytes_to_pixels(data, zero_width, encoders.simple_encode)

        assert False, "bytes2pixels erroneously succeeds on zero width"

    except ValueError:
        pass


def test_bytes2pixels_returns_max_width_on_excessive_width_input():
    data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
    max_width = len(data) // 2
    excessive_width = max_width + 1
    pixels = convert_bytes_to_pixels(data, excessive_width,
                                     encoders.simple_encode)

    assert pixels.width != excessive_width, \
        "{} == {} == {}".format(pixels.width, excessive_width)


### pixels2bytes tests ###

def test_pixels2bytes_returns_correct_length():
    data = bytearray.fromhex('ffff6d06cb0cc436103c2741895c1f5f4861')
    correct_length = len(data)
    width = 3
    test_png_data = convert_bytes_to_pixels(data, width, encoders.simple_encode)
    words = convert_pixels_to_bytes(test_png_data, encoders.simple_decode)

    assert len(words) == correct_length, \
        "{} != {}".format(len(words), correct_length)



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'


def simple_encode(word):
    """Given two bytes interpreted as a 16-bit signed int, if the value is
    positive, assign the first byte to the red channel, and the second, to the
    green one, then set the blue channel to zero. If the value is negative, take
    the two's complement prior to channel assignment and set the blue channel to
    255.

    :param word: two bytes representing a 16-bit signed int
    :type word: bytes()

    :return: three integers between 0 and 255, inclusive
    :rtype: tuple()
    """
    if word[0] & 2**7:
        r = word[1] ^ 255
        g = (word[0] ^ 255) + 1
        b = 255
    else:
        r, g = word
        b = 0

    return r, g, b


def simple_decode(r, g, b):
    """Given three one-byte integers--r, g, and b, respectively--if b != 255,
    return r and g, concatenated in a bytearray. Otherwise, if b == 255, treat
    r and g as a signed, 16-bit integer, taking the little-endian two's
    complement and returning the result.

    :param r: an integer between 0 and 255
    :type r: int
    :param g: an integer between 0 and 255
    :type g: int
    :param b: an integer between 0 and 255
    :type b: int

    :return: a word (two bytes)
    :rtype: bytes()
    """
    if b == 255:
        word = bytearray([(g ^ 255) + 1])
        word += bytearray([r ^ 255])
    else:
        word = bytearray([r, g])

    return word



from setuptools import setup

setup(
    name='transmedia',
    version=__import__('transmedia').__version__,
    description='utilities for thumbing one\'s nose at intellectual property \
    law (or not)',
    url='https://github.com/jonnystorm/transmedia',
    license='WTFPLv2',
    author='jstorm',
    author_email='the.jonathan.storm@gmail.com',
    setup_requires=['nose>=1.0'],
    include_package_data=True,
    packages=['transmedia', 'transmedia.tests'],
    long_description="""transmedia is a collection of utilities for toying\
    with audio/image conversion (and copyright/trademark law).""",
    classifiers=[
        'License :: Public Domain',  # OSI Rejected :: WTFPLv2
        'Programming Language :: Python',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Other Audience',  # anti-IP zealots, the curious
        'Topic :: Multimedia',
    ],
    keywords='media conversion audio image copyright trademark hogwash',
    install_requires=['setuptools'],
)



#-*- coding: utf-8 -*-
#
# Copyright © 2014 Jonathan Storm <the.jonathan.storm@gmail.com>
# This work is free. You can redistribute it and/or modify it under the
# terms of the Do What The Fuck You Want To Public License, Version 2,
# as published by Sam Hocevar. See the COPYING.WTFPL file for more details.

__author__ = 'jstorm'

import math
import os


def integer_side_len_of_smallest_square_containing_n_units(n):
    return int(math.ceil(math.sqrt(n)))


def calculate_output_png_width_from_file(filename):
    bytes_per_word = 2
    word_count = os.path.getsize(filename) / bytes_per_word

    return integer_side_len_of_smallest_square_containing_n_units(word_count)


class PixelArray():
    """Value object with some convenient behavior and properties for rows of
    flat pixels.
    """
    def __init__(self, pixel_rows, byte_depth=3):
        """Initialize new PixelArray object.

        :param pixel_rows: an array of rows of flat pixels
        :type pixel_rows: iterable containing one or more list[int]
        """
        self._byte_depth = byte_depth
        self._pixel_rows = pixel_rows
        self._height = len(pixel_rows)
        self._width = len(max(pixel_rows, key=len)) // self._byte_depth

    def __iter__(self):
        for r in self._pixel_rows:
            # Fill in any missing pixels with black (0, 0, 0)
            pad = [0 for i in range((self._width * self._byte_depth) - len(r))]
            yield r + pad

    @property
    def byte_depth(self):
        return self._byte_depth

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    @property
    def rows(self):
        # Return rows padded with zeros
        return [r for r in self]

    @property
    def rows_raw(self):
        # Return unpadded rows
        return self._pixel_rows
